﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace ForzaStudio
{
    public enum EndianType
    {
        BigEndian,
        LittleEndian
    }

    public class EndianStream : Stream
    {
        #region Fields
        private Stream @base;
        private BinaryReader reader;
        private BinaryWriter writer;
        private EndianType endian;
        #endregion

        #region Properties
        public override long Position { get { return @base.Position; } set { @base.Position = value; } }
        public override bool CanRead { get { return @base.CanRead; } }
        public override bool CanSeek { get { return @base.CanSeek; } }
        public override bool CanWrite { get { return @base.CanWrite; } }
        public override bool CanTimeout { get { return @base.CanTimeout; } }
        #endregion

        #region Constructor / Destructor
        public EndianStream(Stream stream, EndianType endian)
        {
            @base = stream;
            if (stream.CanRead) reader = new BinaryReader(stream);
            if (stream.CanWrite) writer = new BinaryWriter(stream);
            this.endian = endian;
        }

        public EndianStream(byte[] data, EndianType endian)
        {
            @base = new MemoryStream(data);
            if (@base.CanRead) reader = new BinaryReader(@base);
            if (@base.CanWrite) writer = new BinaryWriter(@base);
            this.endian = endian;
        }

        ~EndianStream()
        {
            Dispose();
        }

        public new void Dispose()
        {
            if (@base != null) @base.Dispose();
            if (reader != null) reader.Close();
            if (writer != null) writer.Close();
        }
        #endregion

        #region Methods
        public override void Flush() { @base.Flush(); }
        public override long Seek(long offset, SeekOrigin origin) { return @base.Seek(offset, origin); }
        public override long Length { get { return @base.Length; } }
        public override void SetLength(long value) { @base.SetLength(value); }
        public override int Read(byte[] buffer, int offset, int count) { return @base.Read(buffer, offset, count); }
        public override void Write(byte[] buffer, int offset, int count) { @base.Write(buffer, offset, count); }

        public bool ReadBoolean()
        {
            return reader.ReadBoolean();
        }

        public byte ReadByte()
        {
            try
            {
                return reader.ReadByte();
            }
            catch (Exception ex)
            {
                return 0;
            }
        }

        public char ReadChar()
        {
            return reader.ReadChar();
        }

        public short ReadInt16()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(2);
                Array.Reverse(data);
                return BitConverter.ToInt16(data, 0);
            }
            else return reader.ReadInt16();
        }

        public ushort ReadUInt16()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(2);
                Array.Reverse(data);
                return BitConverter.ToUInt16(data, 0);
            }
            else return reader.ReadUInt16();
        }

        public int ReadInt32()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(4);
                Array.Reverse(data);
                return BitConverter.ToInt32(data, 0);
            }
            else return reader.ReadInt32();
        }

        public uint ReadUInt32()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(4);
                Array.Reverse(data);
                return BitConverter.ToUInt32(data, 0);
            }
            else return reader.ReadUInt32();
        }

        public long ReadInt64()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(8);
                Array.Reverse(data);
                return BitConverter.ToInt64(data, 0);
            }
            else return reader.ReadInt64();
        }

        public ulong ReadUInt64()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(8);
                Array.Reverse(data);
                return BitConverter.ToUInt64(data, 0);
            }
            else return reader.ReadUInt64();
        }

        public float ReadSingle()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(4);
                Array.Reverse(data);
                return BitConverter.ToSingle(data, 0);
            }
            else return reader.ReadSingle();
        }

        public double ReadDouble()
        {
            if (endian == EndianType.BigEndian)
            {
                byte[] data = reader.ReadBytes(8);
                Array.Reverse(data);
                return BitConverter.ToDouble(data, 0);
            }
            else return reader.ReadDouble();
        }

        public string ReadASCII(int length)
        {
            return ASCIIEncoding.ASCII.GetString(reader.ReadBytes(length));
        }
        public string ReadString()
        {
            return reader.ReadString();
        }

        public byte[] ReadBytes(int count)
        {
            return reader.ReadBytes(count);
        }

        public void Write(byte[] data)
        {
            writer.Write(data);
        }

        public void Write(object obj)
        {
            switch (Convert.GetTypeCode(obj))
            {
                case TypeCode.Boolean:
                case TypeCode.Byte:
                case TypeCode.Char:
                    writer.Write((byte)obj); break;
                case TypeCode.Int16:
                case TypeCode.UInt16:
                    if (endian == EndianType.BigEndian)
                    {
                        byte[] data = BitConverter.GetBytes((ushort)obj);
                        Array.Reverse(data);
                        writer.Write(data);
                    }
                    else writer.Write((ushort)obj);
                    break;
                case TypeCode.Int32:
                case TypeCode.UInt32:
                    if (endian == EndianType.BigEndian)
                    {
                        byte[] data = BitConverter.GetBytes((uint)obj);
                        Array.Reverse(data);
                        writer.Write(data);
                    }
                    else writer.Write((uint)obj);
                    break;
                case TypeCode.Int64:
                case TypeCode.UInt64:
                    if (endian == EndianType.BigEndian)
                    {
                        byte[] data = BitConverter.GetBytes((ulong)obj);
                        Array.Reverse(data);
                        writer.Write(data);
                    }
                    else writer.Write((ulong)obj);
                    break;
                case TypeCode.Single:
                    if (endian == EndianType.BigEndian)
                    {
                        byte[] data = BitConverter.GetBytes((float)obj);
                        Array.Reverse(data);
                        writer.Write(data);
                    }
                    else writer.Write((float)obj);
                    break;
                case TypeCode.Double:
                    if (endian == EndianType.BigEndian)
                    {
                        byte[] data = BitConverter.GetBytes((double)obj);
                        Array.Reverse(data);
                        writer.Write(data);
                    }
                    else writer.Write((double)obj);
                    break;
                case TypeCode.String: writer.Write(ASCIIEncoding.ASCII.GetBytes((string)obj + "\0")); break;    // assumes youre writing an ascii string
                case TypeCode.Object:
                    byte[] bytes = obj as byte[]; // tries converting unknown object to byte array
                    if (bytes != null) writer.Write(bytes);
                    else throw new NotSupportedException("Invalid datatype.");
                    break;
                default: throw new NotSupportedException("Invalid datatype.");
            }
        }
        #endregion
    }
}
