/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Text;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Runtime.InteropServices;

namespace ForzaStudio
{
    #region Enums
    public enum CompressionType : short
    {
        Stored = 0,
        Shrunk,
        Reduce1,
        Reduce2,
        Reduce3,
        Reduce4,
        Implode,
        Token,
        Deflate,
        Deflate64,
        LZX = 21
    }

    public enum XMemCodecType
    {
        Default = 0,
        LZX = 1
    }

    public struct XMemCodecParametersLZX
    {
        public int Flags;
        public int WindowSize;
        public int CompressionPartitionSize;
    }
    #endregion

    public class ForzaArchive : IDisposable
    {
        #region Fields
        Stream io;
        List<ZipFile> files = new List<ZipFile>();
        #endregion

        #region Properties
        public ZipFile this[string path]
        {
            get
            {
                foreach (ZipFile entry in files)
                {
                    if (entry.FileName.Equals(path, StringComparison.InvariantCultureIgnoreCase))
                    {
                        return entry;
                    }
                }
                return null;
            }
        }

        public List<ZipFile> Files
        {
            get { return files; }
            set { files = value; }
        }

        private Stream Io
        {
            get { return io; }
            set { io = value; }
        }
        #endregion

        #region Constructors / Destructor
        public ForzaArchive(string file)
        {
            // open the archive and leave the stream open for future use
            Initialize(new FileStream(file, FileMode.Open, FileAccess.Read, FileShare.Read));
        }

        public ForzaArchive(Stream stream)
        {
            Initialize(stream);
        }

        ~ForzaArchive()
        {
            Dispose();
        }
        #endregion

        #region Methods
        private void Initialize(Stream stream)
        {
            io = stream;
            BinaryReader br = new BinaryReader(io);

            // Read it to make sure its a PkZip file
            io.Seek((int)io.Length - 0x16, SeekOrigin.Begin);
            if (br.ReadInt32() != 0x06054B50)
                throw new InvalidDataException("This is not a valid Xbox 360 archive.");

            // Lets read the structure
            io.Seek((int)io.Length - 0x16, SeekOrigin.Begin);
            EndLocator el = new EndLocator(io);

            // Now that we have this info lets read our directory entries
            io.Seek(el.DirectoryOffset, SeekOrigin.Begin);
            for (int i = 0; i < el.EntriesInDirectory; i++)
            {
                files.Add(new ZipFile(io));
            }
        }

        public void Dispose()
        {
            if (io != null) io.Dispose();
            files = null;
        }

        public static byte[] GetFileData(string archivePath, string filePath)
        {
            try
            {
                using (ForzaArchive archive = new ForzaArchive(archivePath))
                {
                    return archive[filePath].Data;
                }
            }
            catch (Exception ex) { return null; }
        }
        #endregion

        #region Classes
        protected class EndLocator
        {
            #region Fields
            Stream stream;
            string comment;
            short commentLength;
            internal int DirectoryOffset;
            int directorySize;
            short diskNumber;
            internal short EntriesInDirectory;
            short entriesOnDisk;
            int signature;
            short startDiskNumber;
            #endregion

            #region Constructor
            public EndLocator(Stream stream)
            {
                this.stream = stream;
                BinaryReader br = new BinaryReader(stream);
                signature = br.ReadInt32();
                diskNumber = br.ReadInt16();
                startDiskNumber = br.ReadInt16();
                entriesOnDisk = br.ReadInt16();
                EntriesInDirectory = br.ReadInt16();
                directorySize = br.ReadInt32();
                DirectoryOffset = br.ReadInt32();
                commentLength = br.ReadInt16();
                comment = ASCIIEncoding.ASCII.GetString(br.ReadBytes(commentLength));
            }
            #endregion

            #region Methods
            public void Write(Stream ew)
            {
                BinaryWriter bw = new BinaryWriter(stream);
                bw.Write(signature);
                bw.Write(diskNumber);
                bw.Write(startDiskNumber);
                bw.Write(entriesOnDisk);
                bw.Write(EntriesInDirectory);
                bw.Write(directorySize);
                bw.Write(DirectoryOffset);
                bw.Write(commentLength);
                bw.Write(comment);
            }
            #endregion
        }
        #endregion
    }

    public class ZipFile
    {
        #region Fields
        Stream stream;
        int compressedSize;
        CompressionType compression;
        uint crc;
        short diskNumberStart;
        int externalAttributes;
        string extraField;
        short extraFieldLength;
        short fileCommentLength;
        short fileDate;
        string fileName;
        short fileNameLength;
        short fileTime;
        short flags;
        int headerOffset;
        short internalAttributes;
        int signature;
        int uncompressedSize;
        short versionMadeBy;
        short versionToExtract;
        int offset;
        #endregion

        #region Properties
        public string FileName
        {
            get { return fileName; }
            set
            {
                fileName = value;
                fileNameLength = (short)fileName.Length;
            }
        }

        // todo: possibly cache data returned in a private data field?
        public byte[] Data
        {
            get
            {
                // Move to the record
                stream.Seek(headerOffset, SeekOrigin.Begin);

                // Read the record
                Info record = new Info(stream);

                // Decompress our data
                return record.DecompressData();
            }
        }
        #endregion

        #region Constructor
        public ZipFile(Stream stream)
        {
            this.stream = stream;
            BinaryReader br = new BinaryReader(stream);
            offset = (int)stream.Position;
            signature = br.ReadInt32();
            versionMadeBy = br.ReadInt16();
            versionToExtract = br.ReadInt16();
            flags = br.ReadInt16();
            compression = (CompressionType)br.ReadInt16();
            fileTime = br.ReadInt16();
            fileDate = br.ReadInt16();
            crc = br.ReadUInt32();
            compressedSize = br.ReadInt32();
            uncompressedSize = br.ReadInt32();
            fileNameLength = br.ReadInt16();
            extraFieldLength = br.ReadInt16();
            fileCommentLength = br.ReadInt16();
            diskNumberStart = br.ReadInt16();
            internalAttributes = br.ReadInt16();
            externalAttributes = br.ReadInt32();
            headerOffset = br.ReadInt32();
            fileName = ASCIIEncoding.ASCII.GetString(br.ReadBytes(fileNameLength));
            extraField = ASCIIEncoding.ASCII.GetString(br.ReadBytes(extraFieldLength));
        }
        #endregion

        #region Methods
        public void SaveAs(string filename)
        {
            using (FileStream fs = new FileStream(filename, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                byte[] data = Data;
                fs.Write(data, 0, data.Length);
            }
        }
        #endregion

        #region Classes
        protected class Info
        {
            #region Imports
            [DllImport("xcompress.dll", EntryPoint = "XMemCreateDecompressionContext")]
            public static extern int XMemCreateDecompressionContext(
                XMemCodecType codecType,
                int pCodecParams,
                int flags, ref int pContext);

            [DllImport("xcompress.dll", EntryPoint = "XMemDestroyDecompressionContext")]
            public static extern void XMemDestroyDecompressionContext(int context);

            [DllImport("xcompress.dll", EntryPoint = "XMemResetDecompressionContext")]
            public static extern int XMemResetDecompressionContext(int context);

            [DllImport("xcompress.dll", EntryPoint = "XMemDecompressStream")]
            public static extern int XMemDecompressStream(int context,
                byte[] pDestination, ref int pDestSize,
                byte[] pSource, ref int pSrcSize);

            [DllImport("xcompress.dll", EntryPoint = "XMemCreateCompressionContext")]
            public static extern int XMemCreateCompressionContext(
                XMemCodecType codecType, int pCodecParams,
                int flags, ref int pContext);

            [DllImport("xcompress.dll", EntryPoint = "XMemDestroyCompressionContext")]
            public static extern void XMemDestroyCompressionContext(int context);

            [DllImport("xcompress.dll", EntryPoint = "XMemResetCompressionContext")]
            public static extern int XMemResetCompressionContext(int context);

            [DllImport("xcompress.dll", EntryPoint = "XMemCompressStream")]
            public static extern int XMemCompressStream(int context,
                byte[] pDestination, ref int pDestSize,
                byte[] pSource, ref int pSrcSize);
            #endregion

            #region Fields
            private Stream stream;
            internal int CompressedSize;
            CompressionType compression;
            internal uint Crc;
            string extraField;
            short extraFieldLength;
            short fileDate;
            string fileName;
            short fileNameLength;
            short fileTime;
            short flags;
            int signature;
            internal int UncompressedSize;
            short version;
            byte[] data;
            #endregion

            #region Constructor
            public Info(Stream stream)
            {
                this.stream = stream;
                BinaryReader br = new BinaryReader(stream);
                signature = br.ReadInt32();
                version = br.ReadInt16();
                flags = br.ReadInt16();
                compression = (CompressionType)br.ReadInt16();
                fileTime = br.ReadInt16();
                fileDate = br.ReadInt16();
                Crc = br.ReadUInt32();
                CompressedSize = br.ReadInt32();
                UncompressedSize = br.ReadInt32();
                fileNameLength = br.ReadInt16();
                extraFieldLength = br.ReadInt16();
                fileName = ASCIIEncoding.ASCII.GetString(br.ReadBytes(fileNameLength));
                extraField = ASCIIEncoding.ASCII.GetString(br.ReadBytes(extraFieldLength));
                data = br.ReadBytes(CompressedSize);
            }
            #endregion

            #region Methods
            public byte[] DecompressData()
            {
                // Decompress our data
                byte[] buffer;
                switch (compression)
                {
                    case CompressionType.Stored:
                        {
                            return data;
                        }
                    case CompressionType.Deflate:
                        {
                            MemoryStream ms = new MemoryStream(data);
                            DeflateStream ds = new DeflateStream(ms, CompressionMode.Decompress);

                            buffer = new byte[UncompressedSize];
                            if (ds.Read(buffer, 0, UncompressedSize) != UncompressedSize)
                                throw new InvalidDataException("Decompression failed.");

                            break;
                        }
                    case CompressionType.LZX:
                        {
                            //Create our decompression context
                            int decompressionContext = 0;
                            XMemCreateDecompressionContext(
                                XMemCodecType.LZX,
                                0, 0, ref decompressionContext);

                            //Reset our context first
                            XMemResetDecompressionContext(decompressionContext);

                            //Now lets read and decompress
                            buffer = new byte[UncompressedSize];
                            XMemDecompressStream(decompressionContext,
                                buffer, ref UncompressedSize,
                                data, ref CompressedSize);

                            //Go ahead and destory our context
                            XMemDestroyDecompressionContext(decompressionContext);

                            break;
                        }
                    default:
                        {
                            throw new NotSupportedException(string.Format("Compression type of {0} is currently not supported.", compression));
                        }
                }

                // Check our CRC32 now
                if (ComputeCRC32(buffer) != Crc)
                    throw new InvalidDataException("Invalid CRC detected.");

                // Return our data since we are good
                return buffer;
            }

            private static uint ComputeCRC32(byte[] bytes)
            {
                uint[] table = new uint[256];
                for (uint i = 0; i < table.Length; ++i)
                {
                    uint temp = i;
                    for (int j = 8; j > 0; --j)
                    {
                        if ((temp & 1) == 1)
                            temp = ((temp >> 1) ^ 0xedb88320);
                        else
                            temp >>= 1;
                    }
                    table[i] = temp;
                }

                uint crc = 0xffffffff;
                for (int i = 0; i < bytes.Length; ++i)
                {
                    byte index = (byte)(((crc) & 0xff) ^ bytes[i]);
                    crc = ((crc >> 8) ^ table[index]);
                }
                return ~crc;
            }
            #endregion
        }
        #endregion
    }
}