﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace ForzaStudio
{
    public class ForzaCarSection
    {
        #region Fields
        public EndianStream Stream;
        public string Name;
        public ForzaCarBin Parent;
        public ForzaCarSubSection[] SubSections;
        #endregion

        #region Constructor
        public ForzaCarSection(ForzaCarBin car)
        {
            Parent = car;
            Stream = car.Stream;

            if (Stream.Position + 100 > Stream.Length)
            {
                throw new Exception();
            }

            uint unkType = Stream.ReadUInt32();
            if (unkType != 2 && unkType != 5)
            {
                Stream.Position -= 4;
            }

            Vector3 Offset = new Vector3(Stream.ReadSingle(), Stream.ReadSingle(), Stream.ReadSingle());
            Vector3 targetMin = new Vector3(Stream.ReadSingle(), Stream.ReadSingle(), Stream.ReadSingle());
            Vector3 targetMax = new Vector3(Stream.ReadSingle(), Stream.ReadSingle(), Stream.ReadSingle());

            // damage permutations?
            Stream.Position += 28;
            Stream.Seek(Stream.ReadUInt32() * 16, SeekOrigin.Current);
            Stream.Position += 4;
            Stream.Seek(Stream.ReadUInt32() * 2, SeekOrigin.Current);

            if (car.Version == ForzaVersion.Three)
            {
                Stream.Position += 8;
            }
            else if (car.Version == ForzaVersion.Four)
            {
                Stream.Position += 12;
            }

            Name = Stream.ReadASCII(Stream.ReadByte());
            Stream.Position += 4;

            // parse any extra lod vertex information (lod1-lod5)
            uint lodVertexCount = Stream.ReadUInt32();
            uint lodVertexSize = Stream.ReadUInt32();
            ForzaVertex[] lodVertices = null;
            if (lodVertexCount > 0)
            {
                lodVertices = new ForzaVertex[lodVertexCount];
                for (int i = 0; i < lodVertexCount; i++)
                {
                    lodVertices[i] = new ForzaVertex(car.Version, ForzaVertexType.Car, Stream, lodVertexSize);
                }
            }
            Stream.Position += 4;

            // parse through each piece
            uint subpartCount = Stream.ReadUInt32();
            SubSections = new ForzaCarSubSection[subpartCount];
            for (int i = 0; i < subpartCount; i++)
            {
                SubSections[i] = new ForzaCarSubSection(this);
            }

            // parse through section vertices (lod0)
            Stream.Position += 4;
            int vertexCount = Stream.ReadInt32();
            uint vertexSize = Stream.ReadUInt32();
            ForzaVertex[] lod0Vertices = null;
            if (vertexCount > 0)
            {
                lod0Vertices = new ForzaVertex[vertexCount];
                for (int i = 0; i < vertexCount; i++)
                {
                    lod0Vertices[i] = new ForzaVertex(car.Version, ForzaVertexType.Car, Stream, vertexSize);
                }
            }

            if (car.Version == ForzaVersion.Four)
            {
                Stream.Position += 9;
                Stream.Seek(Stream.ReadUInt32() * Stream.ReadUInt32(), SeekOrigin.Current);
                Stream.Position += 4;
                Stream.Seek(Stream.ReadUInt32() * Stream.ReadUInt32(), SeekOrigin.Current);
            }

            BoundingBox lod0Bounds = BoundingBox.FromVertices(lod0Vertices);
            BoundingBox lodBounds = BoundingBox.FromVertices(lodVertices);

            foreach (ForzaCarSubSection subSection in SubSections)
            {
                subSection.Vertices = Utilities.GenerateVertices(subSection.Lod == 0 ? lod0Vertices : lodVertices, ref subSection.Indices);
                for (int i = 0; i < subSection.Vertices.Length; i++)
                {
                    if (targetMin.Length() > 0 && targetMax.Length() > 0)
                    {
                        if (subSection.Lod == 0)
                        {
                            subSection.Vertices[i].position = Utilities.CalculateBoundTargetValue(subSection.Vertices[i].position, lod0Bounds.Min, lod0Bounds.Max, targetMin, targetMax);
                        }
                        else
                        {
                            subSection.Vertices[i].position = Utilities.CalculateBoundTargetValue(subSection.Vertices[i].position, lodBounds.Min, lodBounds.Max, targetMin, targetMax);
                        }
                    }

                    subSection.Vertices[i].position += Offset;
                    subSection.Vertices[i].texture0.X = subSection.Vertices[i].texture0.X * subSection.XUVScale + subSection.XUVOffset;
                    subSection.Vertices[i].texture0.Y = 1.0f - (subSection.Vertices[i].texture0.Y * subSection.YUVScale + subSection.YUVOffset);
                    subSection.Vertices[i].texture1.X = subSection.Vertices[i].texture1.X * subSection.XUV2Scale + subSection.XUV2Offset;
                    subSection.Vertices[i].texture1.Y = 1.0f - (subSection.Vertices[i].texture1.Y * subSection.YUV2Scale + subSection.YUV2Offset);
                }
                subSection.Indices = Utilities.GenerateTriangleList(subSection.Indices, subSection.FaceCount);
            }
        }
        #endregion
    }
}
