﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;

namespace ForzaStudio
{
    /// <summary>
    /// Base descriptor for various Forza resource files.
    /// </summary>
    public abstract class ForzaResource : IDisposable
    {
        #region Fields

        /// <summary>
        /// Unique identifier for the resource.
        /// </summary>
        public Guid Id = Guid.NewGuid();

        /// <summary>
        /// A collection of additional objects associated with the resource file.
        /// </summary>
        public Dictionary<string, object> Tags = new Dictionary<string, object>();

        #endregion

        #region Properties

        /// <summary>
        /// Identifies which forza version the resource belongs to.
        /// </summary>
        public ForzaVersion Version { get; set; }

        /// <summary>
        /// A path on the local filesystem or within an archive where the resource file is located.
        /// </summary>
        public string FilePath { get; protected set; }

        /// <summary>
        /// The path to the archive where the resource file is located.
        /// </summary>
        public string ArchivePath { get; protected set; }

        /// <summary>
        /// The name of the resource file.
        /// </summary>
        public string FileName
        {
            get
            {
                return Path.GetFileNameWithoutExtension(FilePath).ToLowerInvariant();
            }
        }

        /// <summary>
        /// The extension of the resource file.
        /// </summary>
        public string Extension
        {
            get
            {
                return Path.GetExtension(FilePath).ToLowerInvariant();
            }
        }

        /// <summary>
        /// The friendly name of the resource.
        /// </summary>
        public string Name { get; set; }

        /// <summary>
        /// A description of the resource.
        /// </summary>
        public string Description { get; set; }

        #endregion

        #region Constructor
        protected ForzaResource(string filePath, string archivePath = null)
        {
            if (filePath != null)
            {
                FilePath = filePath.ToLowerInvariant();
                if (archivePath != null)
                {
                    ArchivePath = archivePath.ToLowerInvariant();
                }
            }
            else throw new ArgumentNullException("A file path must be specified");
        }
        #endregion

        #region Methods

        /// <summary>
        /// Uncached access to the resource file's data.
        /// </summary>
        /// <returns></returns>
        public byte[] GetData()
        {
            try
            {
                if (!Utilities.IsNullOrWhiteSpace(ArchivePath) && !Utilities.IsNullOrWhiteSpace(FilePath) && File.Exists(ArchivePath))
                {
                    using (ForzaArchive archive = new ForzaArchive(ArchivePath))
                    {
                        return archive[FilePath].Data;
                    }
                }
                else if (!Utilities.IsNullOrWhiteSpace(FilePath) && File.Exists(FilePath))
                {
                    using (FileStream fs = new FileStream(FilePath, FileMode.Open, FileAccess.Read, FileShare.Read))
                    {
                        byte[] data = new byte[fs.Length];
                        fs.Read(data, 0, data.Length);
                        return data;
                    }
                }
            }
            catch (Exception ex) { }
            return null;
        }

        /// <summary>
        /// Gets streaming access to data loaded in memory.
        /// The caller must dispose of the stream returned or a memory leak will occur.
        /// </summary>
        /// <returns></returns>
        public Stream GetDataStream()
        {
            byte[] data = GetData();
            if (data != null)
            {
                return new MemoryStream(data);
            }
            else return null;
        }

        public abstract void Dispose();
        #endregion
    }
}
