﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Reflection;
using System.Drawing.Imaging;
using System.Globalization;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using XInput = Microsoft.Xna.Framework.Input;

namespace ForzaStudio
{
    #region Enums
    /// <summary>
    /// Forza title version.
    /// </summary>
    public enum ForzaVersion
    {
        Unknown = 0,
        One,
        Two,
        Three,
        Four
    }
    #endregion

    public partial class ForzaStudioForm : Form
    {
        #region Fields
        public Image SelectedTexture;
        public RenderScene Scene;
        public int SelectedVertexCount;
        public int SelectedFaceCount;        
        #endregion

        #region Constructor
        public ForzaStudioForm()
        {
            InitializeComponent();
            BindCameraSpeeds();
            SetDoubleBuffered(pnlTextureDisplay);
            Scene = new RenderScene(ref tvModels, ref pnlViewport);
        }
        #endregion

        #region General Form Events
        private void tabForzaObjects_Selecting(object sender, TabControlCancelEventArgs e)
        {
            bool showViewport = e.TabPageIndex == 0;
            pnlViewport.Visible = showViewport;
            pnlTextureDisplay.Visible = !showViewport;
            pnlTextureDisplay.BackColor = System.Drawing.Color.White;
        }

        // our half-assed render loop :/
        private void ForzaStudioForm_Shown(object sender, EventArgs e)
        {
            while (!this.IsDisposed)
            {
                UpdateScene(false);
                UpdateStatus();
                System.Threading.Thread.Sleep(1);
            }
        }

        #endregion

        #region Menu Events
        private void mnuImport_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog ofd = new OpenFileDialog())
            {
                ofd.Multiselect = false;
                ofd.Filter = "All Forza Files|*.zip;*.carbin;*.rmb.bin;*.xds;*.bix|Forza Archive (.zip)|*.zip|Forza Car Object (.carbin)|*.carbin|Forza Track Object (.rmb.bin)|*.rmb.bin|Forza Texture (.xds, .bix)|*.xds;*.bix";

                if (ofd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    // out with the old
                    tvModels.Nodes.Clear();
                    tvTextures.Nodes.Clear();

                    // in with the new
                    ImportFile(ofd.FileName);

                    // notify the user
                    MessageBox.Show("Import Complete!", "Finished", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }

            tvModels.Sort();
            tvTextures.Sort();

            // reset viewports
            pnlTextureDisplay.Invalidate();
            pnlViewport.Invalidate();

            UpdateStatus();
        }

        /// <summary>
        /// Takes a zip file and extracts it's contents to the specified directory.
        /// Provides a much easier alternative than QuickBMS.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mnuExtractZip_Click(object sender, EventArgs e)
        {
            try
            {
                using (OpenFileDialog ofd = new OpenFileDialog())
                {
                    ofd.Filter = "Forza Archive (.zip)|*.zip";
                    if (ofd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    {
                        using (FolderBrowserDialog fbd = new FolderBrowserDialog())
                        {
                            if (fbd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                            {
                                using (ForzaArchive zip = new ForzaArchive(ofd.FileName))
                                {
                                    foreach (ZipFile file in zip.Files)
                                    {
                                        try
                                        {
                                            UpdateStatus(string.Format("Extracting {0}", file.FileName));
                                            string savePath = fbd.SelectedPath + "\\" + file.FileName;
                                            Directory.CreateDirectory(Path.GetDirectoryName(savePath));
                                            file.SaveAs(savePath);
                                        }
                                        catch (Exception ex)
                                        {
                                            MessageBox.Show(string.Format("Failed to extract {0} from {1}.", file.FileName, Path.GetFileName(ofd.FileName)), "Forza Studio Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                                        }
                                    }
                                }
                                MessageBox.Show("Extraction Complete!", "Finished", MessageBoxButtons.OK, MessageBoxIcon.Information);
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Forza Studio Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            UpdateStatus();
        }

        private void mnuExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void mnuAbout_Click(object sender, EventArgs e)
        {
            StringBuilder about = new StringBuilder();
            about.AppendLine("Developed By:\tMike Davis");
            about.AppendLine("Website:\t\thttp://codeescape.com/");
            about.AppendLine("Research:\thttp://forum.xentax.com/");
            MessageBox.Show(about.ToString(), "About Forza Studio", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }
        #endregion

        #region Texture Events


        private void ctxTextureMenu_Opening(object sender, CancelEventArgs e)
        {
            e.Cancel = tvTextures.Nodes.Count == 0;
        }

        private void ctxTextureTreeMenu_Opening(object sender, CancelEventArgs e)
        {
            e.Cancel = tvTextures.Nodes.Count == 0;
        }

        private void tvTextures_AfterSelect(object sender, TreeViewEventArgs e)
        {
            try
            {
                if (SelectedTexture != null)
                {
                    SelectedTexture.Dispose();
                }
                SelectedTexture = ((ForzaTexture)e.Node.Tag).GetImage();
            }
            catch (Exception ex)
            {
                // todo: display friendly pop-up message
                // failed to parse texture
            }

            // repaint viewport
            pnlTextureDisplay.Invalidate();
        }

        private void pnlTextureDisplay_Paint(object sender, PaintEventArgs e)
        {
            if (tvTextures.SelectedNode != null && SelectedTexture != null)
            {
                try
                {
                    // clamp dimensions
                    int width = Math.Min(e.ClipRectangle.Width, SelectedTexture.Width);
                    int height = Math.Min(e.ClipRectangle.Height, SelectedTexture.Height);

                    // resize preserving the aspect ratio
                    double aspectRatio = Math.Abs(Math.Max((double)SelectedTexture.Width / width, (double)SelectedTexture.Height / height));
                    width = (int)(SelectedTexture.Width / aspectRatio);
                    height = (int)(SelectedTexture.Height / aspectRatio);

                    // center it (give or take a pixel)
                    int xPos = (int)((double)e.ClipRectangle.Width - width) / 2;
                    int yPos = (int)((double)e.ClipRectangle.Height - height) / 2;

                    if (mnuShowTextureAlpha.Checked)
                    {
                        e.Graphics.DrawImage(SelectedTexture, xPos, yPos, width, height);
                    }
                    else
                    {
                        using (MemoryStream ms = new MemoryStream())
                        {
                            SelectedTexture.Save(ms, ImageFormat.Bmp);
                            e.Graphics.DrawImage(Image.FromStream(ms), xPos, yPos, width, height);
                        }
                    }
                }
                catch (Exception ex)
                {
                    // todo: display friendly pop-up message
                    // invalid texture format
                }
            }
        }

        private void pnlTextureDisplay_Resize(object sender, EventArgs e)
        {
            // repaint viewport
            pnlTextureDisplay.Invalidate();
        }

        private void mnuShowTextureAlpha_CheckedChanged(object sender, EventArgs e)
        {
            // repaint viewport
            pnlTextureDisplay.Invalidate();
        }

        private void mnuSaveTextureAs_Click(object sender, EventArgs e)
        {
            try
            {
                // save the current texture to file
                using (SaveFileDialog sfd = new SaveFileDialog())
                {
                    sfd.Filter = "Bitmap (*.bmp)|*.bmp|PNG (*.png)|*.png|JPEG (*.jpg)|*.jpg";
                    if (sfd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    {
                        // get desired format
                        ImageFormat format;
                        switch (Path.GetExtension(sfd.FileName.ToLowerInvariant()))
                        {
                            case ".bmp": format = ImageFormat.Bmp; break;
                            case ".png": format = ImageFormat.Png; break;
                            case ".jpg": format = ImageFormat.Jpeg; break;
                            case ".jpeg": format = ImageFormat.Jpeg; break;
                            default: throw new NotSupportedException();
                        }

                        // save the image
                        using (Image img = ((ForzaTexture)tvTextures.SelectedNode.Tag).GetImage())
                        {
                            img.Save(sfd.FileName, format);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                // todo: show messagebox with friendly error message and then detailed exception information
            }
        }

        // assumes .bmp format to preserve the alpha channel
        private void mnuExtractAllTextures_Click(object sender, EventArgs e)
        {
            using (FolderBrowserDialog fbd = new FolderBrowserDialog())
            {
                if (fbd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    foreach (TreeNode node in tvTextures.Nodes)
                    {
                        ForzaTexture tex = node.Tag as ForzaTexture;
                        UpdateStatus(string.Format("Extracting {0}", tex.FileName));
                        using (Image img = tex.GetImage())
                        {
                            img.Save(fbd.SelectedPath + "\\" + tex.FileName + ".bmp", ImageFormat.Bmp);
                        }
                    }
                    MessageBox.Show("Extraction Complete!", "Finished", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
        }

        // assumes .bmp format to preserve the alpha channel
        private void mnuExtractCheckedTextures_Click(object sender, EventArgs e)
        {
            using (FolderBrowserDialog fbd = new FolderBrowserDialog())
            {
                if (fbd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    foreach (TreeNode node in tvTextures.Nodes)
                    {
                        if (node.Checked)
                        {
                            ForzaTexture tex = node.Tag as ForzaTexture;
                            UpdateStatus(string.Format("Extracting {0}", tex.FileName));
                            using (Image img = tex.GetImage())
                            {
                                img.Save(fbd.SelectedPath + "\\" + tex.FileName + ".bmp", ImageFormat.Bmp);
                            }
                        }
                    }
                    MessageBox.Show("Extraction Complete!", "Finished", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
        }

        #endregion

        #region Model Events

        private void tvModels_AfterCheck(object sender, TreeViewEventArgs e)
        {
            // make sure to unregister event first so children check status changes will remain uninterrupted
            tvModels.AfterCheck -= new TreeViewEventHandler(tvModels_AfterCheck);
            SetChildrenCheckedStatus(e.Node, e.Node.Checked);
            tvModels.AfterCheck += new TreeViewEventHandler(tvModels_AfterCheck);
            pnlViewport.Invalidate();

            SelectedVertexCount = GetSelectedVertexCount(tvModels.Nodes);
            SelectedFaceCount = GetSelectedFaceCount(tvModels.Nodes);
            Application.DoEvents();
        }

        private void mnuShowWireframe_Click(object sender, EventArgs e)
        {
            Scene.ShowWireframe ^= true;
            Scene.Update(true);
        }

        private void mnuExportObj_Click(object sender, EventArgs e)
        {
            ExportWavefrontObj(true);
        }

        private void pnlViewport_Resize(object sender, EventArgs e)
        {
            Scene.ResetGraphicsDevice();
        }

        private void pnlViewport_Paint(object sender, PaintEventArgs e)
        {
            Scene.Update(true);
        }

        private void ctxModelMenu_Opening(object sender, CancelEventArgs e)
        {
            e.Cancel = tvModels.Nodes.Count == 0;
        }

        private void ctxViewportMenu_Opening(object sender, CancelEventArgs e)
        {
            e.Cancel = tvModels.Nodes.Count == 0;
        }

        private void mnuExtractCheckedModels_Click(object sender, EventArgs e)
        {
            ExportWavefrontObj(true);
        }

        private void mnuExtractAllModels_Click(object sender, EventArgs e)
        {
            ExportWavefrontObj(false);
        }

        #endregion

        #region Methods

        void Form1_MouseWheel(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (e.Delta < 0)
                Scene.Camera.ApplyZoomForce(0.02f);
            else if (e.Delta > 0)
                Scene.Camera.ApplyZoomForce(-0.02f);
        }

        private void UpdateScene(bool forceUpdate)
        {
            try
            {
                Scene.Camera.SpeedScale = ((ComboBoxItem<float>)cmbViewportSpeed.SelectedItem).Value;
                //Camera.LookSpeedScale = 1;// ((ComboBoxItem<float>)cmbViewportSpeed.SelectedItem).Value * 1.25f;
                
                MouseState mouse = Mouse.GetState();
                KeyboardState keyboard = Keyboard.GetState();
                GamePadState gamepad = GamePad.GetState(PlayerIndex.One);

                // apply movement
                if (keyboard.IsKeyDown(XInput.Keys.W))
                {
                    Scene.Camera.ApplyForce(Scene.Camera.ForwardDirection);
                }
                if (keyboard.IsKeyDown(XInput.Keys.A))
                {
                    Scene.Camera.ApplyForce(-Scene.Camera.HorizontalDirection);
                }
                if (keyboard.IsKeyDown(XInput.Keys.S))
                {
                    Scene.Camera.ApplyForce(-Scene.Camera.ForwardDirection);
                }
                if (keyboard.IsKeyDown(XInput.Keys.D))
                {
                    Scene.Camera.ApplyForce(Scene.Camera.HorizontalDirection);
                }
                if (keyboard.IsKeyDown(XInput.Keys.Q))
                {
                    Scene.Camera.ApplyForce(new Vector3(0, -1, 0));
                }
                if (keyboard.IsKeyDown(XInput.Keys.E))
                {
                    Scene.Camera.ApplyForce(new Vector3(0, 1, 0));
                }
                if (keyboard.IsKeyDown(XInput.Keys.I))
                {
                    Scene.Camera.ApplyLookForce(new Vector2(0, 0.1f));
                }
                if (keyboard.IsKeyDown(XInput.Keys.J))
                {
                    Scene.Camera.ApplyLookForce(new Vector2(-0.1f, 0));
                }
                if (keyboard.IsKeyDown(XInput.Keys.K))
                {
                    Scene.Camera.ApplyLookForce(new Vector2(0, -0.1f));
                }
                if (keyboard.IsKeyDown(XInput.Keys.L))
                {
                    Scene.Camera.ApplyLookForce(new Vector2(0.1f, 0));
                }

                Scene.Update();
            }
            catch (DeviceLostException ex)
            {
                Scene.ResetGraphicsDevice();
            }
        }

        public static void SetDoubleBuffered(Control control)
        {
            // set instance non-public property with name "DoubleBuffered" to true
            typeof(Control).InvokeMember("DoubleBuffered",
                BindingFlags.SetProperty | BindingFlags.Instance | BindingFlags.NonPublic,
                null, control, new object[] { true });
        }

        private void ImportFile(string fileName)
        {
            try
            {
                SelectedVertexCount = 0;
                SelectedFaceCount = 0;
                if (SelectedTexture != null)
                {
                    SelectedTexture.Dispose();
                    SelectedTexture = null;
                }

                if (fileName.EndsWith(".zip", StringComparison.InvariantCultureIgnoreCase))
                {
                    using (ForzaArchive zip = new ForzaArchive(fileName))
                    {
                        foreach (ZipFile file in zip.Files)
                        {
                            try
                            {
                                string name = file.FileName;
                                if (name.EndsWith(".carbin", StringComparison.InvariantCultureIgnoreCase))
                                {
                                    ImportForzaCarBin(fileName, name);
                                }
                                else if (name.EndsWith(".rmb.bin", StringComparison.InvariantCultureIgnoreCase))
                                {
                                    ImportForzaTrackBin(fileName, name);
                                }
                                else if (name.EndsWith(".xds", StringComparison.InvariantCultureIgnoreCase) ||
                                    (name.EndsWith(".bix", StringComparison.InvariantCultureIgnoreCase) &&
                                    !name.EndsWith("_b.bix", StringComparison.InvariantCultureIgnoreCase)))
                                {
                                    ImportForzaTexture(fileName, name);
                                }
                            }
                            catch (Exception ex)
                            {
                                MessageBox.Show(string.Format("Failed to import {0}\n\n{1}", fileName, ex.ToString()), "Forza Studio Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            }

                            // refresh UI
                            Application.DoEvents();
                        }
                    }
                }
                else if (fileName.EndsWith(".carbin", StringComparison.InvariantCultureIgnoreCase))
                {
                    ImportForzaCarBin(null, fileName);
                }
                else if (fileName.EndsWith(".rmb.bin", StringComparison.InvariantCultureIgnoreCase))
                {
                    ImportForzaTrackBin(null, fileName);
                }
                else if (fileName.EndsWith(".xds", StringComparison.InvariantCultureIgnoreCase) ||
                    fileName.EndsWith(".bix", StringComparison.InvariantCultureIgnoreCase) &&
                    !fileName.EndsWith("_b.bix", StringComparison.InvariantCultureIgnoreCase))
                {
                    ImportForzaTexture(null, fileName);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(string.Format("Failed to import {0}\n\n{1}", fileName, ex.ToString()), "Forza Studio Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            // refresh UI
            Application.DoEvents();
        }

        private void ImportForzaCarBin(string archivePath, string fileName)
        {
            // make sure we're in the right place
            if (!fileName.EndsWith(".carbin", StringComparison.InvariantCultureIgnoreCase))
                throw new NotSupportedException();

            UpdateStatus(string.Format("Importing {0}", fileName));

            // parse the file
            using (ForzaCarBin carBin = new ForzaCarBin(archivePath, fileName))
            {
                // create the required node structure
                TreeNode rootNode = CreateNode(tvModels.Nodes, carBin.FileName);
                foreach (ForzaCarSection section in carBin.Sections)
                {
                    foreach (ForzaCarSubSection subSection in section.SubSections)
                    {
                        TreeNode lodNode = CreateNode(rootNode.Nodes, "LOD" + subSection.Lod);
                        TreeNode sectionNode = CreateNode(lodNode.Nodes, section.Name);
                        TreeNode subSectionNode = new TreeNode(subSection.Name);
                        subSectionNode.Tag = new ForzaMesh(carBin.FileName + "_" + section.Name + "_" + subSection.Name, subSection.Name, subSection.Indices, subSection.Vertices);
                        sectionNode.Nodes.Add(subSectionNode);
                    }
                }
            }
        }

        private void ImportForzaTrackBin(string archivePath, string fileName)
        {
            // make sure we're in the right place
            if (!fileName.EndsWith(".rmb.bin", StringComparison.InvariantCultureIgnoreCase))
                throw new NotSupportedException();

            UpdateStatus(string.Format("Importing {0}", fileName));

            // parse the file
            using (ForzaTrackBin trackBin = new ForzaTrackBin(archivePath, fileName))
            {
                // create the required node structure
                TreeNode rootNode = CreateNode(tvModels.Nodes, trackBin.FileName.Remove(trackBin.FileName.IndexOf('.')));
                foreach (ForzaTrackSection section in trackBin.Sections)
                {
                    TreeNode typeNode = CreateNode(rootNode.Nodes, section.Type);
                    TreeNode sectionNode = CreateNode(typeNode.Nodes, section.Name);
                    foreach (ForzaTrackSubSection subSection in section.SubSections)
                    {
                        TreeNode subSectionNode = new TreeNode(subSection.Name);
                        subSectionNode.Tag = new ForzaMesh(trackBin.FileName + "_" + section.Name + "_" + subSection.Name, subSection.Name, subSection.Indices, subSection.Vertices);
                        sectionNode.Nodes.Add(subSectionNode);
                    }
                }
            }
        }

        private void ImportForzaTexture(string archivePath, string fileName)
        {
            if (!fileName.EndsWith(".xds", StringComparison.InvariantCultureIgnoreCase) &&
                !fileName.EndsWith(".bix", StringComparison.InvariantCultureIgnoreCase) &&
                !fileName.EndsWith("_b.bix", StringComparison.InvariantCultureIgnoreCase))
                throw new NotSupportedException();

            UpdateStatus(string.Format("Importing {0}", fileName));

            ForzaTexture tex = new ForzaTexture(archivePath, fileName);
            TreeNode node = new TreeNode(tex.FileName);
            node.Tag = tex;
            tvTextures.Nodes.Add(node);
        }

        // helper function to create/return a node in a treeview
        private TreeNode CreateNode(TreeNodeCollection nodes, string name)
        {
            TreeNode node = nodes[name];
            if (node == null)
            {
                // node not found, create it
                node = new TreeNode(name);
                node.Name = name;
                nodes.Add(node);
            }
            return node;
        }

        private void SetChildrenCheckedStatus(TreeNode node, bool isChecked)
        {
            foreach (TreeNode n in node.Nodes)
            {
                n.Checked = isChecked;
                SetChildrenCheckedStatus(n, isChecked);
            }
        }

        private void ExportWavefrontObj(bool selective)
        {
            using (SaveFileDialog sfd = new SaveFileDialog())
            {
                sfd.Filter = "Wavefront OBJ (*.obj)|*.obj";

                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    // get models
                    List<ForzaMesh> meshes = GetModelMeshes(tvModels.Nodes, selective);

                    UpdateStatus("Building materials");

                    // build material list
                    Hashtable materials = new Hashtable();
                    foreach (ForzaMesh mesh in meshes)
                    {
                        materials[mesh.MaterialName] = 0;
                    }

                    // create .mtl file
                    if (materials.Count > 0)
                    {
                        using (FileStream mtl = new FileStream(sfd.FileName.Replace(".obj", "") + ".mtl", FileMode.Create, FileAccess.Write, FileShare.Read))
                        using (StreamWriter msw = new StreamWriter(mtl))
                        {
                            foreach (DictionaryEntry entry in materials)
                            {
                                msw.WriteLine("newmtl " + entry.Key);
                                Vector3 color = Utilities.GetRandomColor(entry.Key.GetHashCode()).ToVector3();
                                msw.WriteLine("kd {0} {1} {2}", color.X, color.Y, color.Z);
                            }
                        }
                    }

                    using (FileStream wObj = new FileStream(sfd.FileName, FileMode.Create, FileAccess.Write, FileShare.Read))
                    using (StreamWriter sw = new StreamWriter(wObj))
                    {
                        // write header
                        sw.WriteLine("# Extracted by Forza Studio 4.0 - http://codeescape.com/" + Environment.NewLine);

                        // reference .mtl file
                        if (materials.Count > 0)
                        {
                            sw.WriteLine("mtllib " + Path.GetFileName(sfd.FileName).Replace(".obj", "") + ".mtl");
                        }

                        // export model information
                        int faceBaseIndex = 0;
                        foreach (ForzaMesh mesh in meshes)
                        {
                            UpdateStatus(string.Format("Extracting {0}", mesh.Name));
                            sw.WriteLine("g {0} # {1} vertices {2} faces", mesh.Name, mesh.VertexCount, mesh.FaceCount);
                            foreach (ForzaVertex vertex in mesh.Vertices)
                            {
                                // export coordinate
                                sw.WriteLine("v {0} {1} {2}",
                                    vertex.position.X.ToString(CultureInfo.InvariantCulture),
                                    vertex.position.Y.ToString(CultureInfo.InvariantCulture),
                                    vertex.position.Z.ToString(CultureInfo.InvariantCulture));
                            }

                            foreach (ForzaVertex vertex in mesh.Vertices)
                            {
                                // export uv
                                sw.WriteLine("vt {0} {1}",
                                    vertex.texture0.X.ToString(CultureInfo.InvariantCulture),
                                    vertex.texture0.Y.ToString(CultureInfo.InvariantCulture));
                            }

                            foreach (ForzaVertex vertex in mesh.Vertices)
                            {
                                // export normal
                                sw.WriteLine("vn {0} {1} {2}",
                                    vertex.normal.X.ToString(CultureInfo.InvariantCulture),
                                    vertex.normal.Y.ToString(CultureInfo.InvariantCulture),
                                    vertex.normal.Z.ToString(CultureInfo.InvariantCulture));
                            }

                            sw.WriteLine("usemtl {0}", mesh.MaterialName);
                            for (int j = 0; j < mesh.Indices.Length; j += 3)
                            {
                                int i1 = mesh.Indices[j + 0] + 1;
                                int i2 = mesh.Indices[j + 1] + 1;
                                int i3 = mesh.Indices[j + 2] + 1;

                                // don't bother extracting degenerate (non-planar) triangles
                                if (i1 != i2 && i2 != i3 && i1 != i3)
                                {
                                    sw.WriteLine("f {0}/{0}/{0} {1}/{1}/{1} {2}/{2}/{2}",
                                        faceBaseIndex + mesh.Indices[j + 0] + 1,
                                        faceBaseIndex + mesh.Indices[j + 1] + 1,
                                        faceBaseIndex + mesh.Indices[j + 2] + 1);
                                }
                            }
                            faceBaseIndex += mesh.Vertices.Length;
                        }
                    }
                    MessageBox.Show("Extraction Complete!", "Finished", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            UpdateStatus();
        }

        int GetSelectedVertexCount(TreeNodeCollection nodes)
        {
            int count = 0;
            foreach (TreeNode node in nodes)
            {
                if (node.Checked && node.Tag != null && node.Tag.GetType() == typeof(ForzaMesh))
                {
                    count += ((ForzaMesh)node.Tag).VertexCount;
                }
                count += GetSelectedVertexCount(node.Nodes);
            }
            return count;
        }

        int GetSelectedFaceCount(TreeNodeCollection nodes)
        {
            int count = 0;
            foreach (TreeNode node in nodes)
            {
                if (node.Checked && node.Tag != null && node.Tag.GetType() == typeof(ForzaMesh))
                {
                    count += ((ForzaMesh)node.Tag).FaceCount;
                }
                count += GetSelectedFaceCount(node.Nodes);
            }
            return count;
        }

        List<ForzaMesh> GetModelMeshes(TreeNodeCollection nodes, bool selective)
        {
            List<ForzaMesh> meshes = new List<ForzaMesh>();
            foreach (TreeNode node in nodes)
            {
                if ((!selective || node.Checked) && node.Tag != null && node.Tag.GetType() == typeof(ForzaMesh))
                {
                    meshes.Add(node.Tag as ForzaMesh);
                }
                meshes.AddRange(GetModelMeshes(node.Nodes, selective));
            }
            return meshes;
        }

        private void BindCameraSpeeds()
        {
            cmbViewportSpeed.Items.Clear();
            for (int i = 1; i <= 20; i++)
            {
                cmbViewportSpeed.Items.Add(new ComboBoxItem<float>((float)i, string.Format("{0}x Camera Speed", i)));
            }
            cmbViewportSpeed.SelectedIndex = 0;
        }

        private void UpdateSelectedMeshStatus()
        {
            if (SelectedFaceCount > 0 && SelectedVertexCount > 0)
            {
                lblInformation.Text = string.Format("Displaying {0} vertices and {1} faces", SelectedVertexCount, SelectedFaceCount);
            }
            else lblInformation.Text = string.Empty;
        }

        private void UpdateSelectedTextureStatus()
        {
            if (SelectedTexture != null)
            {
                lblInformation.Text = string.Format("Selected texture dimensions of {0}x{1}", SelectedTexture.Width, SelectedTexture.Height);
            }
            else lblInformation.Text = string.Empty;
        }

        private void UpdateStatus(string text = null)
        {
            if (text == null)
            {
                if (tabForzaObjects.SelectedIndex == 0)
                {
                    UpdateSelectedMeshStatus();
                }
                else
                {
                    UpdateSelectedTextureStatus();
                }
            }
            else
            {
                lblInformation.Text = text;
            }
            Application.DoEvents();
        }
        #endregion
    }
}