﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Drawing.Imaging;
using System.Diagnostics;

namespace ForzaStudio
{
    public class ForzaTexture : ForzaResource
    {
        #region Imports
        [DllImport("xds.dll", EntryPoint = "GetTextureSize")]
        private unsafe static extern void GetTextureSize(byte* inputData, int inputSize, out int outputWidth, out int outputHeight);

        [DllImport("xds.dll", EntryPoint = "Create")]
        private unsafe static extern void Create(byte* inputData, int inputSize, void* outputData);

        [DllImport("xds.dll", EntryPoint = "ConvertToBitmap")]
        private unsafe static extern bool ConvertToBitmap(byte* inputData, int inputSize, int d3dFormat, int width, int height, void* outputData);
        #endregion

        #region Constructors
        // loads a texture from a zipped file
        public ForzaTexture(string archivePath, string filePath)
            : base(filePath, archivePath)
        {
            
        }

        public ForzaTexture(string filePath)
            : base(filePath)
        {

        }

        #endregion

        #region Methods

        // xbox 360 texture container
        private unsafe Image LoadXds(Stream stream)
        {
            int xdsSize = (int)stream.Length;

            // allocate managed byte array large enough to contain aligned xds file
            byte[] xds = new byte[xdsSize + 0x1000 * 2];

            // load xds data
            fixed (byte* xdsPtr = xds)
            {
                // pointer to byte array
                uint basePtr = new UIntPtr(xdsPtr).ToUInt32();

                // offset of 4kb aligned location inside of byte array
                byte* alignedPtr = (byte*)new UIntPtr((basePtr + 0xFFF) & 0xFFFFF000).ToPointer();

                // offset of 4kb aligned location inside of byte array
                int alignedOffset = (int)(new UIntPtr(alignedPtr).ToUInt32() - basePtr);

                // streaming access to managed byte array
                using (MemoryStream ms = new MemoryStream(xds))
                using (BinaryWriter bw = new BinaryWriter(ms))
                {
                    ms.Position = alignedOffset;   // skip to aligned location

                    // construct the first part of the header in big endian format
                    bw.Write((int)0x01000000);
                    bw.Write((int)0x44325854);
                    bw.Write((int)0x30000000);
                    bw.Write((int)0x34000000);
                    bw.Write((int)0x14000000);
                    bw.Write(new byte[28]);

                    // read the other half of the header
                    stream.Read(xds, alignedOffset + 48, 52);

                    // check for valid xds file, although theres no guarantee that another file type won't have these values
                    if (BitConverter.ToInt32(xds, alignedOffset + 48) != 0x03000000 || BitConverter.ToInt32(xds, alignedOffset + 52) != 0x01000000)
                    {
                        throw new Exception("Not a valid XDS file");
                    }

                    // must align texture data on 4kb boundaries
                    stream.Read(xds, alignedOffset + 4096, xdsSize - 52);

                    // pass pre-aligned xds data into xds.dll and get tga bytes
                    int width, height;
                    GetTextureSize(alignedPtr, xdsSize, out width, out height);

                    Bitmap bmp = null;
                    try
                    {
                        // create a new bitmap and lock its position in memory
                        bmp = new Bitmap(width, height, PixelFormat.Format32bppArgb);
                        BitmapData bmpData = bmp.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.ReadWrite, bmp.PixelFormat);

                        // convert the texture and save to the bitmap in memory
                        Create(alignedPtr, xdsSize, bmpData.Scan0.ToPointer());
                        bmp.UnlockBits(bmpData);
                        return bmp;
                    }
                    catch (Exception ex)
                    {
                        if (bmp != null) bmp.Dispose();
                        throw new Exception("Unable to load texture.");
                    }
                }
            }
        }

        // forza-specific texture container
        private unsafe Image LoadBix(byte[] data, int width, int height, int format)
        {
            if (data != null && data.Length > 0 && width > 0 && height > 0 && format > 0)
            {
                Bitmap bmp = null;
                try
                {
                    // create a new bitmap and lock its position in memory
                    bmp = new Bitmap(width, height, PixelFormat.Format32bppArgb);
                    BitmapData bmpData = bmp.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.ReadWrite, bmp.PixelFormat);

                    // perform the conversion
                    fixed (byte* bixPtr = data)
                    {
                        if (!ConvertToBitmap(bixPtr, data.Length, format, width, height, bmpData.Scan0.ToPointer()))
                        {
                            throw new Exception();
                        }

                        bmp.UnlockBits(bmpData);
                        return bmp;
                    }
                }
                catch (Exception ex)
                {
                    if (bmp != null) bmp.Dispose();
                    throw new Exception("Unable to load texture.");
                }
            }
            else throw new ArgumentException();
        }

        /// <summary>
        /// Uncached access to the texture image.
        /// </summary>
        /// <returns></returns>
        public Image GetImage()
        {
            try
            {
                if (FilePath.EndsWith(".xds", StringComparison.InvariantCultureIgnoreCase))
                {
                    using (MemoryStream ms = new MemoryStream(GetData()))
                    {
                        return LoadXds(ms);
                    }
                }
                else if (FilePath.EndsWith(".bix", StringComparison.InvariantCultureIgnoreCase))
                {
                    // get the directory path where the bix file is located
                    string directoryPath = Path.GetDirectoryName(FilePath);

                    // get paths to both versions of bix objects
                    string bixlessPath = FilePath.Replace("_b.bix", string.Empty).Replace(".bix", string.Empty);
                    string bixPath = bixlessPath + ".bix";
                    string bixBPath = bixlessPath + "_b.bix";

                    // load both versions of bix data
                    byte[] bixData = null, bixBData = null;
                    if (Utilities.IsNullOrWhiteSpace(ArchivePath))
                    {
                        // load from the filesystem
                        using (FileStream fs = new FileStream(bixPath, FileMode.Open, FileAccess.Read, FileShare.Read))
                        {
                            bixData = new byte[fs.Length];
                            fs.Read(bixData, 0, bixData.Length);
                        }

                        using (FileStream fs = new FileStream(bixBPath, FileMode.Open, FileAccess.Read, FileShare.Read))
                        {
                            bixBData = new byte[fs.Length];
                            fs.Read(bixBData, 0, bixBData.Length);
                        }
                    }
                    else
                    {
                        // load from the archive
                        bixData = ForzaArchive.GetFileData(ArchivePath, bixPath);
                        bixBData = ForzaArchive.GetFileData(ArchivePath, bixBPath);
                    }

                    // read the 28 byte header
                    uint magic, width, height, levels, format, totalSize, mainSize;
                    using (EndianStream es = new EndianStream(bixData, EndianType.BigEndian))
                    {
                        magic = es.ReadUInt32();
                        if (magic != 0x42495830 && magic != 0x42495831)  // 'BIX0' && 'BIX1'
                        {
                            throw new NotSupportedException("Unrecognized bix file format.");
                        }
                        width = es.ReadUInt32();
                        height = es.ReadUInt32();
                        levels = es.ReadUInt32();    // not confirmed...
                        format = es.ReadUInt32();  // D3DFMT
                        totalSize = es.ReadUInt32(); // bix a and b data size combined
                        mainSize = es.ReadUInt32();  // bix b data size
                    }

                    // check for bix b data file
                    return LoadBix(bixBData, (int)width, (int)height, (int)format);
                }
                else throw new NotSupportedException("Unknown file extension");
            }
            catch (Exception ex) { }

            return null;
        }


        public override void Dispose()
        {
            
        }
        #endregion
    }
}