﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Graphics.PackedVector;


namespace ForzaStudio
{
    public enum ForzaVertexType
    {
        Car,
        Track
    }

    public enum IndexType
    {
        TriList = 4,
        TriStrip = 6
    }

    public struct ForzaVertex
    {
        #region Fields
        public Vector3 position;
        public Vector2 texture0;
        public Vector2 texture1;
        public Vector3 normal;
        public Color color;

        public static VertexElement[] VertexElements =
        {
            new VertexElement(0, 0, VertexElementFormat.Vector3, VertexElementMethod.Default, VertexElementUsage.Position, 0),
            new VertexElement(0, 12, VertexElementFormat.Vector2, VertexElementMethod.Default, VertexElementUsage.TextureCoordinate, 0),
            new VertexElement(0, 20, VertexElementFormat.Vector2, VertexElementMethod.Default, VertexElementUsage.TextureCoordinate, 1),
            new VertexElement(0, 28, VertexElementFormat.Vector3, VertexElementMethod.Default, VertexElementUsage.Normal, 0),
            new VertexElement(0, 40, VertexElementFormat.Color, VertexElementMethod.Default, VertexElementUsage.Color, 0)
        };

        public static int SizeInBytes = 44;
        #endregion

        #region Constructor
        public ForzaVertex(ForzaVersion version, ForzaVertexType type, EndianStream stream, uint size)
        {
            // initial declarations
            this.position = new Vector3();
            this.texture0 = new Vector2();
            this.texture1 = new Vector2();
            this.normal = new Vector3();
            this.color = Color.Gray;

            float x, y, z, s;
            switch (type)
            {
                case ForzaVertexType.Car:
                    #region Car
                    switch (version)
                    {
                        case ForzaVersion.Three:
                            switch (size)
                            {
                                case 0x10:
                                    x = ConvertHalfToFloat(stream.ReadUInt16());
                                    y = ConvertHalfToFloat(stream.ReadUInt16());
                                    z = ConvertHalfToFloat(stream.ReadUInt16());
                                    s = ConvertHalfToFloat(stream.ReadUInt16());
                                    position = new Vector3(x * s, y * s, z * s);
                                    texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                                    normal = ToN3(stream.ReadUInt32());
                                    break;
                                case 0x28:
                                    x = ConvertHalfToFloat(stream.ReadUInt16());
                                    y = ConvertHalfToFloat(stream.ReadUInt16());
                                    z = ConvertHalfToFloat(stream.ReadUInt16());
                                    s = ConvertHalfToFloat(stream.ReadUInt16());
                                    position = new Vector3(x * s, y * s, z * s);
                                    texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                                    texture1 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                                    normal = ToN3(stream.ReadUInt32());
                                    stream.Position += 16;
                                    ToN3(stream.ReadUInt32());  // tangent
                                    break;
                                case 0x2C:
                                    x = ConvertHalfToFloat(stream.ReadUInt16());
                                    y = ConvertHalfToFloat(stream.ReadUInt16());
                                    z = ConvertHalfToFloat(stream.ReadUInt16());
                                    s = ConvertHalfToFloat(stream.ReadUInt16());
                                    position = new Vector3(x * s, y * s, z * s);
                                    texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                                    normal = ToN3(stream.ReadUInt32());
                                    stream.Position += 16;
                                    ToN3(stream.ReadUInt32());  // tangent
                                    stream.Position += 8;
                                    break;
                                default: throw new NotSupportedException();
                            }
                            break;
                        case ForzaVersion.Four:
                            switch (size)
                            {
                                case 0x20:
                                    x = ShortN(stream.ReadInt16());
                                    y = ShortN(stream.ReadInt16());
                                    z = ShortN(stream.ReadInt16());
                                    s = ShortN(stream.ReadInt16());
                                    position = new Vector3(x * s, y * s, z * s);
                                    texture0 = new Vector2(UShortN(stream.ReadUInt16()), UShortN(stream.ReadUInt16()));
                                    texture1 = new Vector2(UShortN(stream.ReadUInt16()), UShortN(stream.ReadUInt16()));
                                    Matrix m = Matrix.CreateFromQuaternion(new Quaternion(ShortN(stream.ReadInt16()), ShortN(stream.ReadInt16()), ShortN(stream.ReadInt16()), ShortN(stream.ReadInt16())));
                                    normal = new Vector3(m.M11, m.M12, m.M13);
                                    new Vector3(m.M21, m.M22, m.M23);  // tangent
                                    stream.Position += 8;
                                    break;
                                default: throw new NotSupportedException();
                            }
                            break;
                        default: throw new NotSupportedException();
                    }
                    break;
                    #endregion
                case ForzaVertexType.Track:
                    #region Track
                    switch (size)
                    {
                        case 0x0C:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            break;
                        case 0x10:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            break;
                        case 0x14:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            break;
                        case 0x18:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            texture1 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            break;
                        case 0x1C:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            texture1 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            stream.Position += 4;
                            break;
                        case 0x20:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            texture1 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            stream.Position += 8;
                            break;
                        case 0x24:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            texture1 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            stream.Position += 12;
                            break;
                        case 0x28:
                            x = stream.ReadSingle();
                            y = stream.ReadSingle();
                            z = stream.ReadSingle();
                            position = new Vector3(x, y, z);
                            normal = GetNormalized101010(stream.ReadUInt32());
                            texture0 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            texture1 = new Vector2(ConvertHalfToFloat(stream.ReadUInt16()), ConvertHalfToFloat(stream.ReadUInt16()));
                            stream.Position += 16;
                            break;
                        default: throw new NotSupportedException();
                    }
                    break;
                    #endregion
                default: throw new NotSupportedException();
            }
        }
        #endregion

        #region Methods

        private float ShortN(short value)
        {
            return (float)value / 0x7FFF;
        }

        private float UShortN(ushort value)
        {
            return (float)value / 0xFFFF;
        }

        private Vector3 GetNormalized101010(uint packedValue)
        {
            Microsoft.Xna.Framework.Graphics.PackedVector.Normalized101010 n = new Microsoft.Xna.Framework.Graphics.PackedVector.Normalized101010();
            n.PackedValue = packedValue;
            return n.ToVector3();
        }

        /// <summary>
        /// Converts a 16-bit packed half-precision float to a 32-bit single precision float.
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private unsafe float ConvertHalfToFloat(uint value)
        {
            uint mantissa = 0;
            uint exponent = 0;
            uint result;

            mantissa = value & 0x3FF;

            if ((value & 0x7C00) != 0)  // The value is normalized
            {
                exponent = (value >> 10) & 0x1F;
            }
            else if (mantissa != 0) // The value is denormalized
            {
                // Normalize the value in the resulting float
                exponent = 1;

                do
                {
                    exponent--;
                    mantissa <<= 1;
                } while ((mantissa & 0x400) == 0);

                mantissa &= 0x3FF;
            }
            else exponent = 0xFFFFFF90; // (uint)-112

            result = ((value & 0x8000) << 16) | ((exponent + 112) << 23) | (mantissa << 13); // sign | exponent | mantissa
            return *(float*)&result;
        }

        Vector3 ToN3(uint u)
        {
            Vector3 v = Vector3.Zero;
            uint[] SignExtendXY = new uint[] { 0, 0xFFFFF800 };
            uint[] SignExtendZ = new uint[] { 0, 0xFFFFFC00 };
            uint element = u & 0x7FF;
            v.X = (float)(short)(element | SignExtendXY[element >> 10]) / 1023.0f;
            element = (u >> 11) & 0x7FF;
            v.Y = (float)(short)(element | SignExtendXY[element >> 10]) / 1023.0f;
            element = (u >> 22) & 0x3FF;
            v.Z = (float)(short)(element | SignExtendZ[element >> 9]) / 511.0f;
            return v;
        }
        #endregion
    }
}
