﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using XInput = Microsoft.Xna.Framework.Input;

namespace ForzaStudio
{
    public class RenderScene : IDisposable
    {
        #region Fields
        private GraphicsDevice Graphics;
        private BasicEffect Ambiance;
        private BasicEffect ColoredVertices;
        private bool GraphicsDisabled = false;
        #endregion

        #region Properties
        public TreeView Models { get; private set; }
        public Panel Viewport { get; private set; }
        public Camera Camera { get; private set; }       
        public bool ShowWireframe { get; set; }
        #endregion

        #region Constructor/Destructor
        /// <summary>
        /// Creates a render scene.
        /// </summary>
        /// <param name="models">The list of candidate models to be rendered.</param>
        /// <param name="viewport">The viewport to render on.</param>
        public RenderScene(ref TreeView models, ref Panel viewport)
        {
            if (models == null || viewport == null)
                throw new ArgumentNullException();

            Viewport = viewport;
            Models = models;
            Camera = new Camera(viewport);

            try
            {
                InitializeGraphicsDevice();
                Update(true);
            }
            catch (Exception ex)
            {
                GraphicsDisabled = true;
                MessageBox.Show("Graphics card incompatibility detected. The model viewport will remain disabled.");
            }
        }

        ~RenderScene()
        {
            Dispose();
        }
        #endregion

        #region Methods

        /// <summary>
        /// Updates the scene.
        /// </summary>
        /// <param name="forceRedraw">Forces the scene to be redrawn regardless of whether or not the camera has moved.</param>
        public void Update(bool forceRedraw = false)
        {
            try
            {
                if (!GraphicsDisabled)
                {
                    Camera.Update(Viewport.Width, Viewport.Height);
                    if (forceRedraw || Camera.HasChanged)
                    {
                        // update views
                        Ambiance.View = Camera.View;
                        Ambiance.Projection = Camera.Projection;
                        Ambiance.World = Camera.World;
                        Ambiance.DirectionalLight0.Direction = Camera.ForwardDirection;
                        ColoredVertices.View = Camera.View;
                        ColoredVertices.Projection = Camera.Projection;
                        ColoredVertices.World = Camera.World;

                        // clear the scene
                        Graphics.Clear(new Microsoft.Xna.Framework.Graphics.Color(153, 180, 209, 255));

                        // draw the grid
                        DrawGrid(10, 1.0f);

                        // draw the model
                        DrawCheckedModels(Models.Nodes, FillMode.Solid);

                        // add in the wireframe overlay if necessary
                        if (ShowWireframe)
                        {
                            DrawCheckedModels(Models.Nodes, FillMode.WireFrame);
                        }

                        // paint the viewport
                        Graphics.Present();
                    }
                }
            }
            catch (Exception ex)
            {

            }
        }

        /// <summary>
        /// Initializes the graphics device.
        /// </summary>
        private void InitializeGraphicsDevice()
        {
            if (!GraphicsDisabled)
            {
                // set up graphics device options
                PresentationParameters presentationParams = new PresentationParameters();
                presentationParams.IsFullScreen = false;
                presentationParams.BackBufferCount = 2;
                //presentationParams.SwapEffect = SwapEffect.Flip;
                presentationParams.BackBufferWidth = Viewport.Width;
                presentationParams.BackBufferHeight = Viewport.Height;
                presentationParams.BackBufferFormat = SurfaceFormat.Unknown;
                presentationParams.EnableAutoDepthStencil = true;
                presentationParams.AutoDepthStencilFormat = DepthFormat.Depth24Stencil8;
                presentationParams.PresentationInterval = PresentInterval.Default;

                // attempt to create the device
                Graphics = new GraphicsDevice(GraphicsAdapter.DefaultAdapter, DeviceType.Hardware, Viewport.Handle, presentationParams);
                Graphics.RenderState.CullMode = CullMode.CullClockwiseFace;

                // creates the ambient effect
                Ambiance = new BasicEffect(Graphics, null);
                Ambiance.Alpha = 1.0f;
                Ambiance.DiffuseColor = new Vector3(0.75f, 0.75f, 0.75f);
                Ambiance.SpecularColor = new Vector3(0.75f, 0.75f, 0.75f);
                Ambiance.SpecularPower = 25.0f;
                Ambiance.AmbientLightColor = new Vector3(0.75f, 0.75f, 0.75f);
                Ambiance.LightingEnabled = true;
                Ambiance.EnableDefaultLighting();
                Ambiance.VertexColorEnabled = true;
                Ambiance.DirectionalLight0.Enabled = true;
                Ambiance.DirectionalLight0.DiffuseColor = Vector3.One;
                Ambiance.DirectionalLight0.SpecularColor = new Vector3(0.75f, 0.75f, 0.75f);
                Ambiance.DirectionalLight1.Enabled = true;
                Ambiance.DirectionalLight1.DiffuseColor = Vector3.One;
                Ambiance.DirectionalLight1.SpecularColor = new Vector3(0.75f, 0.75f, 0.75f);
                Ambiance.DirectionalLight1.Direction = new Vector3(0, -1, 0);

                // creates the effect used for wireframe mode
                ColoredVertices = new BasicEffect(Graphics, null);
                ColoredVertices.Alpha = 1.0f;
                ColoredVertices.DiffuseColor = Vector3.One;
                ColoredVertices.AmbientLightColor = Vector3.One;
                ColoredVertices.VertexColorEnabled = true;
            }
        }

        /// <summary>
        /// Disposes of the scene.
        /// </summary>
        public void Dispose()
        {
            if (Graphics != null) Graphics.Dispose();
            if (Ambiance != null) Ambiance.Dispose();
            if (ColoredVertices != null) ColoredVertices.Dispose();
        }

        /// <summary>
        /// Resets the graphics device in the scene, preserving the current camera position.
        /// </summary>
        public void ResetGraphicsDevice()
        {
            Dispose();
            InitializeGraphicsDevice();
            Update(true);
        }

        private void DrawCheckedModels(TreeNodeCollection nodes, FillMode fill)
        {
            if (fill == FillMode.Solid)
            {
                Ambiance.VertexColorEnabled = true;
                Ambiance.DiffuseColor = new Vector3(0.7f, 0.7f, 0.7f);
                Graphics.RenderState.FillMode = FillMode.Solid;
            }
            else if (fill == FillMode.WireFrame)
            {
                Ambiance.VertexColorEnabled = false;
                Graphics.RenderState.FillMode = FillMode.WireFrame;
                Ambiance.DiffuseColor = new Vector3(0.0f, 0.0f, 0.0f);
            }

            foreach (TreeNode node in nodes)
            {
                DrawCheckedModels(node.Nodes, fill);
                if (node.Checked && node.Tag != null && node.Tag.GetType() == typeof(ForzaMesh))
                {
                    Ambiance.Begin();
                    foreach (EffectPass pass in Ambiance.CurrentTechnique.Passes)
                    {
                        pass.Begin();
                        ForzaMesh mesh = node.Tag as ForzaMesh;
                        mesh.Draw(ref Graphics);
                        pass.End();
                    }
                    Ambiance.End();
                }
            }
        }

        private void DrawGrid(int size, float spacing)
        {
            ColoredVertices.Begin();
            foreach (EffectPass pass in ColoredVertices.CurrentTechnique.Passes)
            {
                pass.Begin();

                float halfSize = (size * spacing) / 2;
                int lineCount = (size + 1) * 4;

                // create vertex buffer
                VertexPositionColor[] vertices = new VertexPositionColor[lineCount + 2];
                for (int i = 0; i < lineCount; i += 4)
                {
                    float index = -halfSize + (i / 4) * spacing;

                    // x axis
                    vertices[i] = new VertexPositionColor(new Vector3(-halfSize, 0, index), Microsoft.Xna.Framework.Graphics.Color.DarkGreen);
                    vertices[i + 1] = new VertexPositionColor(new Vector3(halfSize, 0, index), Microsoft.Xna.Framework.Graphics.Color.White);

                    // y axis
                    vertices[i + 2] = new VertexPositionColor(new Vector3(index, 0, -halfSize), Microsoft.Xna.Framework.Graphics.Color.DarkBlue);
                    vertices[i + 3] = new VertexPositionColor(new Vector3(index, 0, halfSize), Microsoft.Xna.Framework.Graphics.Color.White);
                }

                // create y axis
                vertices[lineCount] = new VertexPositionColor(new Vector3(0, -halfSize, 0), Microsoft.Xna.Framework.Graphics.Color.DarkRed);
                vertices[lineCount + 1] = new VertexPositionColor(new Vector3(0, halfSize, 0), Microsoft.Xna.Framework.Graphics.Color.White);

                Graphics.VertexDeclaration = new VertexDeclaration(Graphics, VertexPositionColor.VertexElements);
                Graphics.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, vertices, 0, vertices.Length / 2);

                pass.End();

            }
            ColoredVertices.End();
        }
        #endregion
    }
}
