﻿/*
    Forza Studio: Forza Motorsport resource extraction tool.

    Copyright (C) 2012 Mike Davis (http://codeescape.com/)

    Permission is hereby granted, free of charge, to any person 
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included 
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using Microsoft.Xna.Framework;

namespace ForzaStudio
{
    public static class Utilities
    {
        public static int CalculateFaceCount(int[] indices, IndexType type)
        {
            if (type == IndexType.TriStrip)
            {
                int j = 0, faceCount = 0;
                while (j < indices.Length - 2)
                {
                    if (indices[j + 2] != -1)
                    {
                        j++;
                        faceCount++;
                    }
                    else j += 3;
                }
                return faceCount;
            }
            else if (type == IndexType.TriList)
            {
                return indices.Length / 3;
            }
            else throw new NotSupportedException("Unknown index type.");
        }

        // get unique indices to determine the vertex count
        // tristrip indices
        public static int CalculateVertexCount(int[] indices)
        {
            Hashtable ht = new Hashtable();
            foreach (int i in indices)
            {
                if (i != -1)
                {
                    ht[i] = 0;
                }
            }
            return ht.Count;
        }

        // compiles a list of vertices used only for this sub part
        public static ForzaVertex[] GenerateVertices(ForzaVertex[] sectionVertices, ref int[] subSectionIndices)
        {
            // build lookup table mapping each section relative index (key) with its piece relative index (value)
            int vertexIndex = 0;
            Hashtable ht = new Hashtable();
            int indicesLength = subSectionIndices.Length;
            for (int i = 0; i < indicesLength; i++)
            {
                int index = subSectionIndices[i];
                if (index != -1)
                {
                    if (ht[index] == null)
                    {
                        ht[index] = vertexIndex;
                        vertexIndex++;
                    }
                    subSectionIndices[i] = (int)ht[index];    // update index
                }
            }

            // pick out piece vertices from parent section
            ForzaVertex[] vertices = new ForzaVertex[vertexIndex];
            foreach (DictionaryEntry entry in ht)
            {
                vertices[(int)entry.Value] = sectionVertices[(int)entry.Key];
            }

            return vertices;
        }

        public static int[] GenerateTriangleList(int[] indices, int faceCount)
        {
            // generate triangle list from strip data
            int[] indexBuffer = new int[faceCount * 3];
            bool windingOrderSwitch = true; // clockwise cull mode
            int f1, f2, f3, f = 0, listIndex = 0;
            while (f < indices.Length - 2)
            {
                f3 = indices[f + 2];
                if (f3 != -1)
                {
                    f1 = indices[f];
                    f2 = indices[f + 1];
                    f++;

                    if (f1 == f2 || f2 == f3 || f1 == f3)
                    {
                        // skip degenerates (non-planar faces)
                    }
                    else if (windingOrderSwitch)
                    {
                        indexBuffer[listIndex] = f1;
                        indexBuffer[listIndex + 1] = f2;
                        indexBuffer[listIndex + 2] = f3;
                    }
                    else
                    {
                        indexBuffer[listIndex] = f2;
                        indexBuffer[listIndex + 1] = f1;
                        indexBuffer[listIndex + 2] = f3;
                    }
                    listIndex += 3;
                    windingOrderSwitch ^= true;
                }
                else
                {
                    windingOrderSwitch = true;
                    f += 3;
                }
            }
            return indexBuffer;
        }

        public static int[] ReadIndices(EndianStream stream, int count, int size)
        {
            int[] indices = new int[count];
            for (int i = 0; i < indices.Length; i++)
            {
                int indexValue = size == 2 ? stream.ReadUInt16() : stream.ReadInt32();
                if ((size == 2 && indexValue == 0xFFFF) || (size == 4 && indexValue >= 0xFFFFFF))
                {
                    indexValue = -1;
                }
                indices[i] = indexValue;
            }
            return indices;
        }

        public static bool IsNullOrWhiteSpace(string s)
        {
            return s == null || s.Trim().Length == 0;
        }

        //public static void AssertEquals<T>(T obj1, T obj2)
        //{
        //    if (!obj1.Equals(obj2))
        //    {
        //        throw new Exception(string.Format("Analyze this! Expected: {0}, Actual: {1}", obj2.ToString(), obj1.ToString()));
        //    }
        //}

        public static void AssertEquals(byte obj1, byte obj2)
        {
            if (!obj1.Equals(obj2))
            {
                throw new Exception(string.Format("Analyze this! Expected: {0}, Actual: {1}", obj2.ToString(), obj1.ToString()));
            }
        }
        public static void AssertEquals(uint obj1, uint obj2)
        {
            if (!obj1.Equals(obj2))
            {
                throw new Exception(string.Format("Analyze this! Expected: {0}, Actual: {1}", obj2.ToString(), obj1.ToString()));
            }
        }
        public static void AssertEquals(float obj1, float obj2)
        {
            if (!obj1.Equals(obj2))
            {
                throw new Exception(string.Format("Analyze this! Expected: {0}, Actual: {1}", obj2.ToString(), obj1.ToString()));
            }
        }

        public static Microsoft.Xna.Framework.Graphics.Color GetRandomColor(int seed)
        {
            Random rand = new Random(seed);
            return new Microsoft.Xna.Framework.Graphics.Color((byte)rand.Next(0, 255), (byte)rand.Next(0, 255), (byte)rand.Next(0, 255));
        }

        public static float CalculateBoundTargetValue(float actualValue, float actualMin, float actualMax, float targetMin, float targetMax)
        {
            if (actualValue > actualMax || actualValue < actualMin)
                throw new ArgumentException("Actual value falls outside of actual range");

            if (targetMin >= targetMax)
                throw new ArgumentException("Invalid target range");

            return targetMin + ((actualValue - actualMin) / (actualMax - actualMin)) * (targetMax - targetMin);
        }

        public static Vector3 CalculateBoundTargetValue(Vector3 actualValue, Vector3 actualMin, Vector3 actualMax, Vector3 targetMin, Vector3 targetMax)
        {
            return targetMin + ((actualValue - actualMin) / (actualMax - actualMin)) * (targetMax - targetMin);
        }
    }

    public struct BoundingBox
    {
        public float MinX;
        public float MaxX;
        public float MinY;
        public float MaxY;
        public float MinZ;
        public float MaxZ;

        public float XRange
        {
            get
            {
                return MaxX - MinX;
            }
        }

        public float YRange
        {
            get
            {
                return MaxY - MinY;
            }
        }

        public float ZRange
        {
            get
            {
                return MaxZ - MaxX;
            }
        }

        public Vector3 Min
        {
            get
            {
                return new Vector3(MinX, MinY, MinZ);
            }
        }

        public Vector3 Max
        {
            get
            {
                return new Vector3(MaxX, MaxY, MaxZ);
            }
        }

        public Vector3 Range
        {
            get
            {
                return new Vector3(MaxX - MinX, MaxY - MinY, MaxZ - MinZ);
            }
        }

        public static BoundingBox FromVertices(ForzaVertex[] vertices)
        {
            BoundingBox box = new BoundingBox();
            if (vertices != null && vertices.Length > 0)
            {
                foreach (ForzaVertex v in vertices)
                {
                    box.MinX = Math.Min(v.position.X, box.MinX);
                    box.MaxX = Math.Max(v.position.X, box.MaxX);
                    box.MinY = Math.Min(v.position.Y, box.MinY);
                    box.MaxY = Math.Max(v.position.Y, box.MaxY);
                    box.MinZ = Math.Min(v.position.Z, box.MinZ);
                    box.MaxZ = Math.Max(v.position.Z, box.MaxZ);
                }
            }
            return box;
        }
    }

    public class ComboBoxItem<T>
    {
        public string Label { get; set; }
        public T Value { get; set; }

        public ComboBoxItem(T value, string label)
        {
            Value = value;
            Label = label;
        }

        public override string ToString()
        {
            return Label ?? string.Empty;
        }
    }

}